from typing import Dict, Union, List

from qiskit import QuantumCircuit
from qiskit.circuit import Qubit, Clbit
from qiskit.result import Result


def parse_result(
        qiskit_result: Result, circuit: QuantumCircuit,
        indexed_results: bool = True) -> Dict[Union[Qubit, int], Dict[str, int]]:
    """
    parse results into a dictionary where keys are the qubits (or its indices) and values are dictionaries containing
    results of states for that qubit (ignoring all other qubits)
    :param qiskit_result: result returned by backend.run(circuit)
    :param circuit: circuit for which the qiskit_result was run
    :param indexed_results: if true keys for dictionary are indices if false it's Qubit objects
    :return: dictionary containing parsed results
    """
    qubit_clbit_mapping = _get_qubit_mapping(circuit)
    parsed_results = {}
    for state, count in qiskit_result.get_counts().items():
        parsed_state = _parse_state(state)
        for qubit_index in range(len(circuit.qubits)):
            key = qubit_index if indexed_results else circuit.qubits[qubit_index]
            measurement_index = qubit_clbit_mapping[qubit_index]

            if measurement_index is None:
                continue

            qubit_state = parsed_state[measurement_index]

            if key not in parsed_results:
                parsed_results[key] = {'0': 0, '1': 0}
            parsed_results[key][qubit_state] += count

    return parsed_results


def _get_qubit_mapping(circuit: QuantumCircuit) -> List[Union[None, int]]:
    """
    return mapping between index of a qubit and a real index of its
    measurement result in the state returned by get_counts(). indices do not consider spaces so state must have spaces
    removed before indices address correctly
    :param circuit: circuit for which to create the mapping
    :return: mapping between qubit indices and its measurement indices
    """
    qubit_mapping = [None] * circuit.num_qubits
    for instruction, qubits, bits in circuit.get_instructions('measure'):
        qubit_index = circuit.qubits.index(qubits[0])
        qubit_mapping[qubit_index] = _get_real_clbit_index(circuit, bits[0])
    return qubit_mapping


def _get_real_clbit_index(circuit: QuantumCircuit, clbit: Clbit) -> int:
    """
    return the index of classical bit in the state returned by counts (indices don't consider spaces
    - they must be removed for the indices to address correctly)
    :param circuit: circuit that clbit is part of
    :param clbit: clbit to get index of
    :return:
    """
    creg_index = 0
    for creg in reversed(circuit.cregs):
        if clbit in creg:
            clbit_index = creg.index(clbit)
            return creg_index + (len(creg) - clbit_index - 1)
        else:
            creg_index += len(creg)


def _parse_state(state: str) -> str:
    """
    removes spaces from str
    :param state: str to parse
    :return: provided str without spaces
    """
    return state.replace(" ", "")



from typing import Sequence, Union, Type, List

from qiskit import QuantumCircuit
from qiskit.circuit import Instruction, Qubit, Clbit
from qiskit.circuit.bit import Bit
from qiskit.circuit.exceptions import CircuitError


def insert_instruction(
        circuit: QuantumCircuit, instruction: Instruction, qubits: Sequence[Union[Qubit, int]],
        clbits: Sequence[Union[Clbit, int]], index: int, in_place: bool = True) -> QuantumCircuit:
    """
    insert instruction at a specified place (in lists of instructions from circuit.data)
    :param circuit: circuit where the instructions should be inserted
    :param instruction: instruction to be inserted
    :param qubits: qubits used for the instruction (can be indexes or objects
    :param clbits: clbits used for the instruction (can be indexes or objects)
    :param index: index where the instruction will be placed in circuit.data
    :param in_place: creates new circuit if False and returns it, otherwise updates the provided circuit and returns it
    :return: circuit with instruction inserted
    """

    new_circuit = circuit if in_place else circuit.copy()

    if not isinstance(instruction, Instruction):
        raise ValueError("specified instruction is not of type Instruction")

    if len(qubits) != instruction.num_qubits or len(clbits) != instruction.num_clbits:
        raise CircuitError(
            "number of qubits or clbits provided doesn't match instruction's qubits and clbits requirements")

    parsed_qubits = _parse_bit(qubits, Qubit, new_circuit)
    parsed_clbits = _parse_bit(clbits, Clbit, new_circuit)
    instruction_tuple = (instruction, parsed_qubits, parsed_clbits)

    new_circuit.data.insert(index, instruction_tuple)
    return new_circuit


def _parse_bit(bits: Sequence[Union[Bit, int]], bit_type: Type[Bit], circuit: QuantumCircuit) -> List[Bit]:
    """
    parse bits into list of bits (not just its indices) or throws exception if incorrect arguments
    :param bits: list of bits or indices of bits to be parsed
    :param bit_type: either Clbit or Qubit
    :param circuit: circuit for which the bits are to be parsed
    :return: list of Clbits/Qubits
    """
    parsed_bits = []
    type_name = {bit_type.__name__}

    for bit in bits:
        if isinstance(bit, bit_type):
            if bit_type == Qubit:
                if bit in circuit.qubits:
                    parsed_bits.append(bit)
                else:
                    raise CircuitError(f"One of the specified {type_name}s is not a part of a circuit, try adding it first")
            elif bit_type == Clbit:
                if bit in circuit.clbits:
                    parsed_bits.append(bit)
                else:
                    raise CircuitError(f"One of the specified {type_name}s is not a part of a circuit, try adding it first")
            else:
                raise ValueError("bit type must be either qiskit.circuit.Clbit or qiskit.circuit.Qubit")
        elif isinstance(bit, int):
            parsed_bits.append(circuit.qubits[bit])
        else:
            raise ValueError(f"Sequence of {type_name}s contains elements that are neither a {type_name} or a int")
    return parsed_bits



from collections.abc import Sequence
from typing import Union

from qiskit import QuantumCircuit
from qiskit.circuit import Instruction, Qubit, Clbit

from qiskit_utils.insert import insert_instruction


class QuantumCircuitEnhanced(QuantumCircuit):
    def insert(
            self, instruction: Instruction, qubits: Union[Sequence[Qubit, int]],
            clbits: Union[Sequence[Clbit, int]], index: int, in_place: bool = True) -> QuantumCircuit:
        """
        insert instruction at a specified place (in lists of instructions from circuit.data)
        :param instruction: instruction to be inserted
        :param qubits: qubits used for the instruction (can be indexes or objects
        :param clbits: clbits used for the instruction (can be indexes or objects)
        :param index: index where the instruction will be placed in self.data
        :param in_place: creates new circuit if False and returns it, otherwise updates self and returns it
        :return: circuit with instruction inserted
        """
        return insert_instruction(self, instruction, qubits, clbits, index, in_place=in_place)



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='qiskit_utils',
    version='1.1.6',
    license='MIT',
    author="Marek Grzesiuk",
    packages=['qiskit_utils'],
    url='https://github.com/mgrzesiuk/qiskit-utils',
    keywords='utility-methods qiskit',
    install_requires=[
          'qiskit>=0.34.2',
      ],
    description="package containing utility methods for qiskit like result parsing and instruction insertion for circuits",
    long_description=long_description,
    long_description_content_type='text/markdown'
)
